/*
 * © NHN Corp. All rights reserved.
 * NHN Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.toast.android.gamebase.sample;

import android.app.Activity;
import android.content.Context;
import android.content.DialogInterface;
import android.graphics.Color;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.toast.android.gamebase.Gamebase;
import com.toast.android.gamebase.GamebaseCallback;
import com.toast.android.gamebase.GamebaseConfiguration;
import com.toast.android.gamebase.GamebaseDataCallback;
import com.toast.android.gamebase.GamebaseWebViewConfiguration;
import com.toast.android.gamebase.GamebaseWebViewStyle;
import com.toast.android.gamebase.analytics.data.GameUserData;
import com.toast.android.gamebase.analytics.data.LevelUpData;
import com.toast.android.gamebase.auth.data.AuthToken;
import com.toast.android.gamebase.auth.data.BanInfo;
import com.toast.android.gamebase.auth.data.GraceBanInfo;
import com.toast.android.gamebase.auth.data.TemporaryWithdrawalInfo;
import com.toast.android.gamebase.auth.mapping.data.ForcingMappingTicket;
import com.toast.android.gamebase.base.GamebaseError;
import com.toast.android.gamebase.base.GamebaseException;
import com.toast.android.gamebase.base.NetworkManager;
import com.toast.android.gamebase.base.ScreenOrientation;
import com.toast.android.gamebase.base.auth.AuthProvider;
import com.toast.android.gamebase.base.auth.AuthProviderCredentialConstants;
import com.toast.android.gamebase.base.auth.AuthProviderProfile;
import com.toast.android.gamebase.base.data.GamebaseDataContainer;
import com.toast.android.gamebase.base.purchase.PurchasableItem;
import com.toast.android.gamebase.base.purchase.PurchasableReceipt;
import com.toast.android.gamebase.base.purchase.PurchaseProvider;
import com.toast.android.gamebase.base.push.PushConfiguration;
import com.toast.android.gamebase.base.push.PushProvider;
import com.toast.android.gamebase.base.push.data.GamebaseNotificationOptions;
import com.toast.android.gamebase.base.push.data.GamebasePushTokenInfo;
import com.toast.android.gamebase.contact.ContactConfiguration;
import com.toast.android.gamebase.error.data.UpdateInfo;
import com.toast.android.gamebase.event.GamebaseEventCategory;
import com.toast.android.gamebase.event.GamebaseEventHandler;
import com.toast.android.gamebase.event.data.GamebaseEventLoggedOutData;
import com.toast.android.gamebase.event.data.GamebaseEventObserverData;
import com.toast.android.gamebase.event.data.GamebaseEventServerPushData;
import com.toast.android.gamebase.event.data.PushAction;
import com.toast.android.gamebase.event.data.PushMessage;
import com.toast.android.gamebase.imagenotice.ImageNoticeConfiguration;
import com.toast.android.gamebase.launching.data.LaunchingInfo;
import com.toast.android.gamebase.launching.data.LaunchingStatus;
import com.toast.android.gamebase.sample.ui.AppFlowHelper;
import com.toast.android.gamebase.sample.ui.MainActivity;
import com.toast.android.gamebase.sample.util.Log;
import com.toast.android.gamebase.terms.data.GamebaseQueryTermsResult;
import com.toast.android.gamebase.terms.data.GamebaseShowTermsViewResult;
import com.toast.android.gamebase.terms.data.GamebaseTermsContent;
import com.toast.android.gamebase.terms.data.GamebaseTermsContentDetail;
import com.toast.android.gamebase.terms.data.GamebaseUpdateTermsConfiguration;
import com.toast.android.gamebase.toastlogger.LoggerConfiguration;
import com.toast.android.gamebase.toastlogger.data.LogEntry;
import com.toast.android.gamebase.toastlogger.data.LogFilter;
import com.toast.android.gamebase.toastlogger.data.LoggerListener;

import org.jetbrains.annotations.NotNull;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class GamebaseManager {
    private static final String TAG = "GamebaseManager";

    // Initialization settings
    // See https://docs.toast.com/en/Game/Gamebase/en/aos-initialization/#initialize
    // NHN Cloud Project ID
    // See https://docs.toast.com/en/TOAST/en/console-guide/#create-projects
    // TODO: [Fix me] Project ID issued from NHN Cloud Project.
    private static final String PROJECT_ID = "6ypq5kwa";
    // See https://docs.toast.com/en/Game/Gamebase/en/oper-app/#client-list
    // TODO: [Fix me] See app client version from Gamebase Console.
    private static final String APP_CLIENT_VERSION = "1.0.0";
    // TODO: [Fix me] Change your store code for initialization.
    private static final String STORE_CODE = PurchaseProvider.StoreCode.GOOGLE;
    // Popup settings
    // See https://docs.toast.com/en/Game/Gamebase/en/aos-initialization/#configuration-settings
    // TODO: [Fix me] CThis value determines whether to use pop-ups provided by Gamebase SDK.
    private static final boolean ENABLE_POPUP = true;
    private static final boolean ENABLE_LAUNCHING_STATUS_POPUP = true;
    private static final boolean ENABLE_BAN_POPUP = true;

    // Debug mode
    // See https://docs.toast.com/en/Game/Gamebase/en/aos-initialization/#debug-mode
    // TODO: [Fix me] Change to 'false' when release mode.
    private static final boolean DEBUG_MODE = true;

    // Gamebase Analytics
    // See https://docs.toast.com/en/Game/Gamebase/en/aos-etc/#analytics
    // TODO: [Fix me] Value for the analytics data of the game.
    private static final int USER_CURRENT_GAME_LEVEL = 10;
    private static final String ANALYTICS_CHANNEL_ID = "channel sample";
    private static final String ANALYTICS_CHARACTER_ID = "character human";

    // Temporary Withdrawal
    // See https://docs.toast.com/en/Game/Gamebase/en/aos-authentication/#temporarywithdrawal
    // TODO: [Fix me] Value of whether to use the temporary withdrawal or not
    private static final boolean USE_TEMPORARY_WITHDRAWAL = false;

    // Grace Ban
    // TODO: [Fix me] Value of whether to use the grace ban or not
    private static final boolean USE_GRACE_BAN = false;

    // Push
    private static PushConfiguration PUSH_CONFIGURATION_CACHE_FROM_TERMS_VIEW = null;

    // GamebaseEventHandler
    private static GamebaseEventHandler GAMEBASE_EVENT_HANDLER = null;

    // Logger
    // See https://docs.toast.com/en/Game/Gamebase/en/aos-logger/#initialize
    // TODO: [Fix me] AppKey for the Log & Crash from NHN Cloud Console > Analytics > Log & Crash Search
    static final String LOG_AND_CRASH_APPKEY = "";

    private static void printGamebaseError(final GamebaseException exception) {
        Log.w(TAG, "Login Exception");
        Log.w(TAG, "Login Exception - Domain : " + exception.getDomain());
        Log.w(TAG, "Login Exception - Message : " + exception.getMessage());
        Log.w(TAG, "Login Exception - Code : " + exception.getCode());
        Log.w(TAG, "Login Exception - Detail : " + exception.toString());
        Log.w(TAG, "--------------------------------------");
    }

    private enum GAME_PLAY_STATUS {
        PLAYABLE,
        STOP_PLAYING,
        INITIALIZE_AGAIN
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Initialization
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void initialize(final Activity activity) {
        Gamebase.setDebugMode(DEBUG_MODE);

        GamebaseConfiguration configuration =
                GamebaseConfiguration.newBuilder(PROJECT_ID, APP_CLIENT_VERSION, STORE_CODE)
                        .enablePopup(ENABLE_POPUP)
                        .enableLaunchingStatusPopup(ENABLE_LAUNCHING_STATUS_POPUP)
                        .enableBanPopup(ENABLE_BAN_POPUP)
                        .build();

        Gamebase.initialize(activity, configuration, new GamebaseDataCallback<LaunchingInfo>() {
            @Override
            public void onCallback(LaunchingInfo launchingInfo, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    GAME_PLAY_STATUS canPlay = GAME_PLAY_STATUS.PLAYABLE;
                    String errorLog = "";
                    final int launchingStatusCode = launchingInfo.getStatus().getCode();
                    switch (launchingStatusCode) {
                        case LaunchingStatus.IN_SERVICE:
                            break;
                        case LaunchingStatus.RECOMMEND_UPDATE:
                            Log.d(TAG, "There is a new version of this application.");
                            break;
                        case LaunchingStatus.IN_SERVICE_BY_QA_WHITE_LIST:
                        case LaunchingStatus.IN_TEST:
                        case LaunchingStatus.IN_REVIEW:
                        case LaunchingStatus.IN_BETA:
                            Log.d(TAG, "You logged in because you are developer.");
                            break;
                        case LaunchingStatus.REQUIRE_UPDATE:
                            canPlay = GAME_PLAY_STATUS.STOP_PLAYING;
                            errorLog = "You have to update this application.";
                            break;
                        case LaunchingStatus.BLOCKED_USER:
                            canPlay = GAME_PLAY_STATUS.INITIALIZE_AGAIN;
                            errorLog = "You are blocked user!";
                            break;
                        case LaunchingStatus.TERMINATED_SERVICE:
                            canPlay = GAME_PLAY_STATUS.INITIALIZE_AGAIN;
                            errorLog = "Game is closed!";
                            break;
                        case LaunchingStatus.INSPECTING_SERVICE:
                        case LaunchingStatus.INSPECTING_ALL_SERVICES:
                            canPlay = GAME_PLAY_STATUS.INITIALIZE_AGAIN;
                            errorLog = "Under maintenance.";
                            break;
                        case LaunchingStatus.INTERNAL_SERVER_ERROR:
                        default:
                            canPlay = GAME_PLAY_STATUS.INITIALIZE_AGAIN;
                            errorLog = "Unknown internal error.";
                            break;
                    }
                    if (canPlay.equals(GAME_PLAY_STATUS.PLAYABLE)) {
                        Log.v(TAG, "Launching Succeeded");
                        moveToShowTermsViewFlow(activity, () -> AppFlowHelper.moveToLoginScene(activity));
                    } else {
                        Log.w(TAG, "Launching Failed(" + launchingStatusCode + ") : " + errorLog);
                        if (canPlay.equals(GAME_PLAY_STATUS.INITIALIZE_AGAIN)) {
                            if (!ENABLE_POPUP || !ENABLE_LAUNCHING_STATUS_POPUP) {
                                AppFlowHelper.showErrorAndReturnToTitle(activity, "Launching Failed", errorLog);
                            } else {
                                AppFlowHelper.returnToTitle(activity);
                            }
                        }
                    }
                } else {
                    UpdateInfo updateInfo = UpdateInfo.from(exception);
                    if (updateInfo != null) {
                        Log.w(TAG, "Launching Exception : " + exception.toJsonString());
                        if (!ENABLE_POPUP || !ENABLE_LAUNCHING_STATUS_POPUP) {
                            // Initialized with not registered game client version.
                            // Show update popup manually.
                            AppFlowHelper.showUnregisteredVersionAndMoveToStore(activity, updateInfo.installUrl, updateInfo.message);
                        }
                        return;
                    }

                    Log.w(TAG, "Launching Exception : " + exception.toJsonString());

                    AppFlowHelper.showErrorAndReturnToTitle(activity,
                            "Launching Exception",
                            exception.toJsonString());
                }
            }
        });
    }

    private static void moveToShowTermsViewFlow(final Activity activity, final Callback callback) {
        showTermsView(activity, (container, exception) -> {
            if (Gamebase.isSuccess(exception)) {
                GamebaseShowTermsViewResult termsViewResult = GamebaseShowTermsViewResult.from(container);
                Log.d(TAG, "GamebaseShowTermsViewResult : " + termsViewResult.toString());
                if (termsViewResult.isTermsUIOpened) {
                    AppFlowHelper.savePushConfiguration(
                            activity.getApplicationContext(), termsViewResult.pushConfiguration);
                }
                if (callback != null) {
                    callback.onCallback();
                }
            } else {
                Log.w(TAG, "showTermsView() failed : " + exception.toString());
                new Thread(() -> {
                    try {
                        Thread.sleep(500);
                    } catch (Exception ignored) {}
                    moveToShowTermsViewFlow(activity, callback);
                }).start();
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Authentication
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void lastProviderLogin(final Activity activity, final OnLoginWithLastestLoginTypeCallback callback) {
        final String lastLoggedInProvider = Gamebase.getLastLoggedInProvider();
        final boolean hasGamebaseAccessToken = Gamebase.getAccessToken() != null;
        Log.d(TAG, "Last Logged in Provider : " + lastLoggedInProvider);

        Gamebase.loginForLastLoggedInProvider(activity, new GamebaseDataCallback<AuthToken>() {
            @Override
            public void onCallback(AuthToken result, GamebaseException exception) {
                boolean triggerCallback = true;
                if (Gamebase.isSuccess(exception)) {
                    onLoginWithLastLoggedInProviderSucceeded(activity, result);
                } else {
                    if (isNetworkError(exception)) {
                        triggerCallback = false;
                        AppFlowHelper.showError(activity, "Network Error", "Check your network.");
                        retryWithInterval(new Runnable() {
                            @Override
                            public void run() {
                                lastProviderLogin(activity, callback);
                            }
                        }, 2);
                    } else if (isBannedUser(exception)) {
                        // Do nothing because you set the 'enableBanPopup' true.
                        // Gamebase will show ban-popup automatically.

                        // Obtaining Ban Information
                        BanInfo banInfo = BanInfo.from(exception);
                        printBanInfo(banInfo);
                    } else {
                        if (hasGamebaseAccessToken && lastLoggedInProvider != null) {
                            Map<String, Object> additionalInfo = new HashMap<>();
                            if (lastLoggedInProvider.equals(AuthProvider.LINE)) {
                                AppFlowHelper.showLineRegionDialog(activity, new AppFlowHelper.Callback() {
                                    @Override
                                    public void onDialogComplete(String region) {
                                        if (region != null && !region.isEmpty()) {
                                            Map<String, Object> additionalInfo = new HashMap<>();
                                            additionalInfo.put(AuthProviderCredentialConstants.LINE_CHANNEL_REGION, region);
                                            GamebaseManager.loginWithIdP(activity, AuthProvider.LINE, additionalInfo);
                                        }
                                    }
                                });
                            } else {
                                loginWithIdP(activity, lastLoggedInProvider, additionalInfo);
                            }
                        } else {
                            // Do nothing. User should select IDP from Game UI.
                        }
                    }
                }

                if (triggerCallback && callback != null) {
                    callback.onCallback();
                }
            }
        });
    }

    public interface OnLoginWithLastestLoginTypeCallback {
        void onCallback();
    }

    public static void loginWithIdP(final Activity activity, final String provider, final Map<String, Object> additionalInfo) {
        Log.d(TAG, "Logger with provider : " + provider);
        Gamebase.login(activity, provider, additionalInfo, new GamebaseDataCallback<AuthToken>() {
            @Override
            public void onCallback(AuthToken result, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    onLoginWithIdPSucceeded(activity, provider, result);
                } else {
                    if (isNetworkError(exception)) {
                        AppFlowHelper.showError(activity, "Network Error", "Check your network.");
                        retryWithInterval(new Runnable() {
                            @Override
                            public void run() {
                                loginWithIdP(activity, provider, additionalInfo);
                            }
                        }, 2);
                    } else if (isBannedUser(exception)) {
                        // Do nothing because you set the 'enableBanPopup' true.
                        // Gamebase will show ban-popup automatically.

                        // Obtaining Ban Information
                        BanInfo banInfo = BanInfo.from(exception);
                        printBanInfo(banInfo);
                    } else {
                        onLoginFailed(activity, exception);
                    }
                }
            }
        });
    }

    private static boolean isNetworkError(final GamebaseException exeption) {
        final int errorCode = exeption.getCode();
        if (errorCode == GamebaseError.SOCKET_ERROR ||
                errorCode == GamebaseError.SOCKET_RESPONSE_TIMEOUT) {
            return true;
        }
        return false;
    }

    private static boolean isBannedUser(final GamebaseException exeption) {
        if (exeption.getCode() == GamebaseError.BANNED_MEMBER) {
            return true;
        }
        return false;
    }

    private static void printBanInfo(final BanInfo banInfo) {
        Log.d(TAG, "Ban Info");
        Log.d(TAG, "Ban Info - User ID : " + banInfo.userId);
        Log.d(TAG, "Ban Info - Ban Type : " + banInfo.banType);
        Log.d(TAG, "Ban Info - Message : " + banInfo.message);
        Log.d(TAG, "Ban Info - Begin Date : " + banInfo.beginDate);
        Log.d(TAG, "Ban Info - End Date : " + banInfo.endDate);
        Log.d(TAG, "--------------------------------------");
    }

    private static void retryWithInterval(final Runnable runnable, final int intervalSecond) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                try {
                    Thread.sleep(intervalSecond * 1000);
                    runnable.run();
                } catch (InterruptedException e) {
                }
            }
        }).start();
    }

    private static void onLoginFailed(final Activity activity, GamebaseException exception) {
        printGamebaseError(exception);
        AppFlowHelper.showError(activity, "Login Failed!", exception.toJsonString());
    }

    public static void onLoginSucceeded(final Activity activity, AuthToken authToken) {
        ((MainActivity) activity).enableLastProviderLogin(false);

        Log.i(TAG, "Gamebase User Id : " + authToken.getUserId());
        Log.i(TAG, "Gamebase Access Token : " + authToken.getAccessToken());
        Log.i(TAG, "LastLoggedInProvider : " + Gamebase.getLastLoggedInProvider());
        Log.d(TAG, "--------------------------------------");

        // Handles All of Gamebase events.
        addGamebaseEventHandler(activity);

        // Initialize Gamebase Analytics
        initializeAnalytics(USER_CURRENT_GAME_LEVEL, ANALYTICS_CHANNEL_ID, ANALYTICS_CHARACTER_ID);

        moveToTemporaryWithdrawFlow(activity, USE_TEMPORARY_WITHDRAWAL, authToken.getTemporaryWithdrawalInfo(), () ->
                moveToGraceBanFlow(activity, USE_GRACE_BAN, authToken.getGraceBanInfo(), () ->
                        moveToRetryTransactionFlow(activity, () ->
                                moveToRegisterPushFlow(activity, () ->
                                        moveToImageNoticeFlow(activity, () -> {
                                            // Start game.
                                            AppFlowHelper.startGame(activity);
                                        })))));
    }

    private static void moveToTemporaryWithdrawFlow(final Activity activity,
                                                    final boolean useTemporaryWithdrawal,
                                                    final TemporaryWithdrawalInfo temporaryWithdrawalInfo,
                                                    final Callback callback) {
        if (useTemporaryWithdrawal && temporaryWithdrawalInfo != null) {
            // This user requested to withdraw game.
            // Ask the user whether to proceed with withdrawal, cancel withdrawal, or continue playing the game.
            AppFlowHelper.showTemporaryWithdrawUI(activity, temporaryWithdrawalInfo, new Callback() {
                @Override
                public void onCallback() {
                    if (callback != null) callback.onCallback();
                }
            });
        } else {
            if (callback != null) callback.onCallback();
        }
    }

    private static void moveToGraceBanFlow(final Activity activity,
                                           final boolean useGraceBan,
                                           final GraceBanInfo graceBanInfo,
                                           final Callback callback) {
        if (useGraceBan && graceBanInfo != null) {
            // 'Grace Ban' feature.
            // This user should payment more.
            // Otherwise, this user will be banned after the specified period.
            // Show the user how much more he have to pay, etc.
            AppFlowHelper.showGraceBanUI(activity, graceBanInfo, () -> {
                if (callback != null) callback.onCallback();
            });
        } else {
            if (callback != null) callback.onCallback();
        }
    }

    private static void moveToRetryTransactionFlow(final Activity activity, final Callback callback) {
        if (hasPurchaseAdapterInProject() && activity != null) {
            // Retry transaction and check not consumed item
            requestNotConsumedItems(activity, new GamebaseDataCallback<List<PurchasableReceipt>>() {
                @Override
                public void onCallback(List<PurchasableReceipt> purchasableReceipts, GamebaseException notComsumeListException) {
                    if (Gamebase.isSuccess(notComsumeListException)) {
                        if (purchasableReceipts != null && purchasableReceipts.size() > 0) {
                            // There are purchased items that require consume.
                            // Game client needs to send a request to the game server to consume,
                            // so that items can be provided.
                        }
                    }
                    if (callback != null) callback.onCallback();
                }
            });
        } else {
            if (callback != null) callback.onCallback();
        }
    }

    private static void moveToRegisterPushFlow(final Activity activity, final Callback callback) {
        PushConfiguration pushConfiguration =
                AppFlowHelper.loadPushConfiguration(activity.getApplicationContext());
        //noinspection ConstantConditions
        if (hasPushAdapterInProject() &&
                activity != null &&
                pushConfiguration != null) {
            registerPush(activity, pushConfiguration, new GamebaseCallback() {
                @Override
                public void onCallback(GamebaseException e) {
                    if (callback != null) callback.onCallback();
                }
            });
        } else {
            if (callback != null) callback.onCallback();
        }
    }

    private static void moveToImageNoticeFlow(final Activity activity, final Callback callback) {
        if (activity != null) {
            showImageNotices(activity, new GamebaseCallback() {
                @Override
                public void onCallback(GamebaseException ignored) {
                    if (callback != null) callback.onCallback();
                }
            });
        } else {
            if (callback != null) callback.onCallback();
        }
    }

    private static void onLoginWithLastLoggedInProviderSucceeded(final Activity activity, AuthToken result) {
        Log.d(TAG, "Login with Last Logged In Provider Success");
        Log.i(TAG, "LastLoggedInProvider : " + Gamebase.getLastLoggedInProvider());

        onLoginSucceeded(activity, result);
    }

    private static void onLoginWithIdPSucceeded(final Activity activity, String provider, AuthToken result) {
        Log.d(TAG, "Login with IdP(" + provider + ") Success");
        Log.i(TAG, provider + " User Id : " + Gamebase.getAuthProviderUserID(provider));
        Log.i(TAG, provider + " Access Token : " + Gamebase.getAuthProviderAccessToken(provider));
        AuthProviderProfile profile = Gamebase.getAuthProviderProfile(provider);
        Log.i(TAG, provider + " Profile : " + (profile != null ? profile.toJsonString() : "null"));

        onLoginSucceeded(activity, result);
    }

    public static boolean isLoggedIn() {
        final String userId = Gamebase.getUserID();
        if (userId == null || userId.equalsIgnoreCase("")) {
            return false;
        }
        return true;
    }

    public static void logout(final Activity activity) {
        Gamebase.logout(activity, new GamebaseCallback() {
            @Override
            public void onCallback(GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    AppFlowHelper.returnToLoginScene(activity);
                } else {
                    AppFlowHelper.showError(activity, "Logout Failed !", exception.toJsonString());
                }
            }
        });
    }

    public static void withdraw(final Activity activity) {
        Gamebase.withdraw(activity, new GamebaseCallback() {
            @Override
            public void onCallback(GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    AppFlowHelper.returnToLoginScene(activity);
                } else {
                    AppFlowHelper.showError(activity, "Withdraw Failed !", exception.toJsonString());
                }
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Image Notice
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void showImageNotices(final Activity activity, final GamebaseCallback onCloseCallback) {
        ImageNoticeConfiguration configuration = ImageNoticeConfiguration.newBuilder()
                .build();
        Gamebase.ImageNotice.showImageNotices(activity, configuration,
                new GamebaseCallback() {
                    @Override
                    public void onCallback(GamebaseException exception) {
                        onCloseCallback.onCallback(exception);
                    }
                }, new GamebaseDataCallback<String>() {
                    @Override
                    public void onCallback(String payload, GamebaseException exception) {
                        if (Gamebase.isSuccess(exception)) {
                            Log.i(TAG, "Clicked Image Notice Payload: " + payload);

                            if (payload.equalsIgnoreCase("mygame://some_custom_scheme")) {
                                // Do something with your custom scheme
                            }
                        }
                    }
                });
    }

    public static void closeImageNotices(final Activity activity) {
        Gamebase.ImageNotice.closeImageNotices(activity);
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Image Notice - Example Code
    //
    ////////////////////////////////////////////////////////////////////////////////

    public static void imageNoticesExample(final Activity activity, final GamebaseCallback onCloseCallback) {
        ImageNoticeConfiguration configuration = ImageNoticeConfiguration.newBuilder()
                .setBackgroundColor("#80000000")
                .setTimeout(5000L)
                .enableAutoCloseByCustomScheme(false)
                .build();
        Gamebase.ImageNotice.showImageNotices(activity, configuration,
                new GamebaseCallback() {
                    @Override
                    public void onCallback(GamebaseException exception) {
                        onCloseCallback.onCallback(exception);
                    }
                }, new GamebaseDataCallback<String>() {
                    @Override
                    public void onCallback(String payload, GamebaseException exception) {
                        if (Gamebase.isSuccess(exception)) {
                            Log.i(TAG, "Clicked Image Notice Payload: " + payload);

                            if (payload.equalsIgnoreCase("GO_TO_IN_GAME_STORE")) {
                                AppFlowHelper.moveToInGameStore(activity);
                                closeImageNotices(activity);
                            } else if (payload.equalsIgnoreCase("GO_TO_PUSH_SETTINGS")) {
                                AppFlowHelper.moveToPushSettingsMenu(activity);
                                closeImageNotices(activity);
                            }
                        }
                    }
                });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Account Mapping
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void addMapping(final Activity activity,
                                  final String providerName,
                                  final GamebaseDataCallback<AuthToken> callback) {
        Gamebase.addMapping(activity, providerName, null, (data, exception) -> {
            if (Gamebase.isSuccess(exception)) {
                Log.d(TAG, "Add Mapping Success");
                Log.i(TAG, "User Id : " + data.getUserId());
                Log.i(TAG, "Gamebase Access Token : " + data.getAccessToken());
                Log.d(TAG, "--------------------------------------");
            } else {
                if (exception.getCode() == GamebaseError.SOCKET_RESPONSE_TIMEOUT ||
                        exception.getCode() == GamebaseError.SOCKET_ERROR ||
                        exception.getCode() == GamebaseError.SOCKET_UNKNOWN_ERROR) {
                    new Thread(() -> {
                        try {
                            Thread.sleep(2000);
                            addMapping(activity, providerName, callback);
                        } catch (Exception e) {
                            if (callback != null) callback.onCallback(data, exception);
                        }
                    }).start();
                    return;
                } else if (exception.getCode() == GamebaseError.AUTH_ADD_MAPPING_ALREADY_MAPPED_TO_OTHER_MEMBER) {
                    AppFlowHelper.moveToAddMappingForciblyFlow(activity, exception,
                            (authToken, addMappingForciblyException) -> {
                                if (!Gamebase.isSuccess(addMappingForciblyException)) {
                                    printGamebaseError(addMappingForciblyException);
                                    AppFlowHelper.showError(activity, "Add Mapping Forcibly Failed!",
                                            addMappingForciblyException.toJsonString());
                                }
                                if (callback != null) callback.onCallback(authToken, addMappingForciblyException);
                            });
                    return;
                }
                printGamebaseError(exception);
                AppFlowHelper.showError(activity, "Add Mapping Failed!", exception.toJsonString());
            }
            if (callback != null) callback.onCallback(data, exception);
        });
    }

    public static void addMappingForcibly(final Activity activity,
                                          final ForcingMappingTicket forcingMappingTicket,
                                          final GamebaseDataCallback<AuthToken> callback) {
        Gamebase.addMappingForcibly(activity, forcingMappingTicket, (authToken, exception) -> {
            if (!Gamebase.isSuccess(exception)) {
                printGamebaseError(exception);
                AppFlowHelper.showError(activity, "Add Mapping Forcibly Failed!", exception.toJsonString());
            }
            callback.onCallback(authToken, exception);
        });
    }

    public static void changeLogin(final Activity activity,
                                   final ForcingMappingTicket forcingMappingTicket,
                                   final GamebaseDataCallback<AuthToken> callback) {
        Gamebase.changeLogin(activity, forcingMappingTicket, (authToken, exception) -> {
            if (!Gamebase.isSuccess(exception)) {
                printGamebaseError(exception);
                AppFlowHelper.showError(activity, "Change Login Failed!", exception.toJsonString());
            }
            callback.onCallback(authToken, exception);
        });
    }

    public static void removeMapping(final Activity activity,
                                     final String providerName,
                                     final GamebaseCallback callback) {
        Gamebase.removeMapping(activity, providerName, exception -> {
            if (!Gamebase.isSuccess(exception)) {
                printGamebaseError(exception);
                AppFlowHelper.showError(activity, "Remove Mapping Failed!", exception.toJsonString());
            }
            callback.onCallback(exception);
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Purchase
    //
    ////////////////////////////////////////////////////////////////////////////////
    private static boolean hasPurchaseAdapterInProject() {
        List<String> purchaseProviders = new ArrayList<>();
        purchaseProviders.add(PurchaseProvider.StoreCode.GOOGLE);
        purchaseProviders.add(PurchaseProvider.StoreCode.ONESTORE);
        for (String provider : purchaseProviders) {
            try {
                final String lowerProvider = provider.toLowerCase();
                Class.forName("com.toast.android.gamebase.purchase." + lowerProvider +
                        ".Purchase" + lowerProvider.substring(0, 1).toUpperCase() + lowerProvider.substring(1));
                return true;
            } catch (Exception ignored) {
            }
        }
        return false;
    }

    public static void requestNotConsumedItems(final Activity activity, final GamebaseDataCallback<List<PurchasableReceipt>> callback) {
        Gamebase.Purchase.requestItemListOfNotConsumed(activity, new GamebaseDataCallback<List<PurchasableReceipt>>() {
            @Override
            public void onCallback(List<PurchasableReceipt> purchasableReceipts, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    // Succeeded.
                } else {
                    // Failed.
                    Log.w(TAG, "Request purchases failed : " + exception.toString());
                }

                if (callback != null) {
                    callback.onCallback(purchasableReceipts, exception);
                }
            }
        });
    }

    public static void requestSubscriptions(final Activity activity, final GamebaseDataCallback<List<PurchasableReceipt>> callback) {
        Gamebase.Purchase.requestActivatedPurchases(activity, new GamebaseDataCallback<List<PurchasableReceipt>>() {
            @Override
            public void onCallback(List<PurchasableReceipt> purchasableReceipts, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    // Succeeded.
                } else {
                    // Failed.
                    Log.w(TAG, "Request purchases failed : " + exception.toString());
                }

                if (callback != null) {
                    callback.onCallback(purchasableReceipts, exception);
                }
            }
        });
    }

    public static void requestItemList(final Activity activity, final GamebaseDataCallback<List<PurchasableItem>> callback) {
        Gamebase.Purchase.requestItemListPurchasable(activity, new GamebaseDataCallback<List<PurchasableItem>>() {
            @Override
            public void onCallback(List<PurchasableItem> purchasableItems, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    // Succeeded.
                } else {
                    // Failed.
                    Log.w(TAG, "Request item list failed : " + exception.toString());
                }

                if (callback != null) {
                    callback.onCallback(purchasableItems, exception);
                }
            }
        });
    }

    public static void requestPurchase(final Activity activity, final String gamebaseProductId, final GamebaseDataCallback<PurchasableReceipt> callback) {
        Gamebase.Purchase.requestPurchase(activity, gamebaseProductId, new GamebaseDataCallback<PurchasableReceipt>() {
            @Override
            public void onCallback(PurchasableReceipt data, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    // Succeeded.
                } else if (exception.getCode() == GamebaseError.PURCHASE_USER_CANCELED) {
                    // User canceled.
                } else {
                    // To Purchase Item Failed cause of the error
                }

                if (callback != null) {
                    callback.onCallback(data, exception);
                }
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Push
    //
    ////////////////////////////////////////////////////////////////////////////////
    private static boolean hasPushAdapterInProject() {
        List<String> pushProviders = new ArrayList<>();
        pushProviders.add(PushProvider.Type.FCM);
        for (String provider : pushProviders) {
            try {
                final String lowerProvider = provider.toLowerCase();
                Class.forName("com.toast.android.gamebase.push." + lowerProvider +
                        ".Push" + lowerProvider.substring(0, 1).toUpperCase() + lowerProvider.substring(1));
                return true;
            } catch (Exception ignored) {
            }
        }
        return false;
    }

    public static void queryTokenInfo(@NonNull final Activity activity,
                                      @Nullable final GamebaseDataCallback<GamebasePushTokenInfo> callback) {
        Gamebase.Push.queryTokenInfo(activity, new GamebaseDataCallback<GamebasePushTokenInfo>() {
            @Override
            public void onCallback(GamebasePushTokenInfo tokenInfo, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    Log.d(TAG, "QueryTokenInfo Success");
                    Log.i(TAG, "push token : " + tokenInfo.token);
                    Log.i(TAG, "enablePush : " + tokenInfo.agreement.pushEnabled);
                    Log.i(TAG, "enableAdPush : " + tokenInfo.agreement.adAgreement);
                    Log.i(TAG, "enableAdNightPush : " + tokenInfo.agreement.adAgreementNight);
                    Log.d(TAG, "--------------------------------------");
                } else {
                    Log.e(TAG, "QueryTokenInfo failed : " + exception.toString());
                }
                if (callback != null) {
                    callback.onCallback(tokenInfo, exception);
                }
            }
        });
    }

    public static void registerPush(final Activity activity,
                                    final PushConfiguration configuration,
                                    @Nullable final GamebaseCallback callback) {
        Gamebase.Push.registerPush(activity, configuration, exception -> {
            if (Gamebase.isSuccess(exception)) {
                Log.v(TAG, "Register push succeeded");
            } else {
                Log.e(TAG, "Register push failed : " + exception.toString());
            }
            if (callback != null) {
                callback.onCallback(exception);
            }
        });
    }

    public static void enablePushForeground(final Activity activity,
                                            final boolean enableForeground,
                                            @Nullable final GamebaseCallback callback) {
        // Call queryTokenInfo() API to get current push configuration.
        queryTokenInfo(activity, new GamebaseDataCallback<GamebasePushTokenInfo>() {
            @Override
            public void onCallback(GamebasePushTokenInfo tokenInfo, GamebaseException exception) {
                if (!Gamebase.isSuccess(exception)) {
                    if (callback != null) {
                        callback.onCallback(exception);
                    }
                    return;
                }

                final PushConfiguration configuration = PushConfiguration.newBuilder(
                        tokenInfo.agreement.pushEnabled,
                        tokenInfo.agreement.adAgreement,
                        tokenInfo.agreement.adAgreementNight)
                        .build();

                // Call getNotificationOptions() API to get current notification options.
                final GamebaseNotificationOptions currentOptions = Gamebase.Push.getNotificationOptions(activity);
                final GamebaseNotificationOptions notificationOptions = GamebaseNotificationOptions.newBuilder(currentOptions)
                        .enableForeground(enableForeground)
                        .build();

                Gamebase.Push.registerPush(activity, configuration, notificationOptions, new GamebaseCallback() {
                    @Override
                    public void onCallback(GamebaseException exception) {
                        if (!Gamebase.isSuccess(exception)) {
                            Log.e(TAG, "Register push failed : " + exception.toString());
                        }
                        if (callback != null) {
                            callback.onCallback(exception);
                        }
                    }
                });
            }
        });
    }

    public static void setNotificationOptions(final Activity activity,
                                              final boolean enableForeground,
                                              final boolean enableBadge,
                                              final int priority,
                                              @Nullable final String smallIconName,
                                              @Nullable final String soundFileName,
                                              @Nullable final GamebaseCallback callback) {
        // Call queryTokenInfo to get current push configuration.
        queryTokenInfo(activity, new GamebaseDataCallback<GamebasePushTokenInfo>() {
            @Override
            public void onCallback(GamebasePushTokenInfo tokenInfo, GamebaseException exception) {
                if (!Gamebase.isSuccess(exception)) {
                    if (callback != null) {
                        callback.onCallback(exception);
                    }
                    return;
                }

                final PushConfiguration configuration = PushConfiguration.newBuilder(
                        tokenInfo.agreement.pushEnabled,
                        tokenInfo.agreement.adAgreement,
                        tokenInfo.agreement.adAgreementNight)
                        .build();

                // Call getNotificationOptions() API to get current notification options.
                final GamebaseNotificationOptions currentOptions = Gamebase.Push.getNotificationOptions(activity);
                final GamebaseNotificationOptions notificationOptions = GamebaseNotificationOptions.newBuilder(currentOptions)
                        .enableForeground(enableForeground)
                        .enableBadge(enableBadge)
                        // NotificationCompat.PRIORITY_MIN     : -2
                        // NotificationCompat.PRIORITY_LOW     : -1
                        // NotificationCompat.PRIORITY_DEFAULT :  0
                        // NotificationCompat.PRIORITY_HIGH    :  1
                        // NotificationCompat.PRIORITY_MAX     :  2
                        .setPriority(priority)
                        .setSmallIconName(smallIconName)
                        .setSoundFileName(soundFileName)
                        .build();

                Gamebase.Push.registerPush(activity, configuration, notificationOptions, new GamebaseCallback() {
                    @Override
                    public void onCallback(GamebaseException exception) {
                        if (!Gamebase.isSuccess(exception)) {
                            Log.e(TAG, "Register push failed : " + exception.toString());
                        }
                        if (callback != null) {
                            callback.onCallback(exception);
                        }
                    }
                });
            }
        });
    }

    public static void queryNotificationAllowed(@NonNull final Activity activity,
                                                @Nullable final GamebaseDataCallback<Boolean> callback) {
        Gamebase.Push.queryNotificationAllowed(activity, (notificationAllowed, exception) -> {
            if (Gamebase.isSuccess(exception)) {
                Log.d(TAG, "QueryNotificationAllowed Success");
                if (notificationAllowed) {
                    Log.i(TAG, "Notification : Allowed");
                } else {
                    Log.i(TAG, "Notification : Not Allowed");
                }
            } else {
                Log.e(TAG, "QueryNotificationAllowed failed : " + exception.toString());
            }
            if (callback != null) {
                callback.onCallback(notificationAllowed, exception);
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // WebView
    //
    ////////////////////////////////////////////////////////////////////////////////
    // Show WebView.
    public static void showWebView(final Activity activity, final String urlString) {
        Gamebase.WebView.showWebView(activity, urlString);
    }

    public static void showWebView(final Activity activity,
                                   final String url,
                                   final GamebaseWebViewConfiguration configuration) {
        Gamebase.WebView.showWebView(activity, url, configuration);
    }

    public static void showWebView(final Activity activity,
                                   final String url,
                                   final GamebaseWebViewConfiguration configuration,
                                   final GamebaseCallback onCloseCallback,
                                   final List<String> schemeList,
                                   final GamebaseDataCallback<String> onEvent) {
        Gamebase.WebView.showWebView(activity, url, configuration, onCloseCallback, schemeList, onEvent);
    }

    // Close currently displayed WebView.
    public static void closeWebView(final Activity activity) {
        Gamebase.WebView.closeWebView(activity);
    }

    // Open with External Web Browser.
    public static void openExternalBrowser(final Activity activity, final String urlString) {
        Gamebase.WebView.openWebBrowser(activity, urlString);
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // WebView - Example Code
    //
    ////////////////////////////////////////////////////////////////////////////////
    private static void webviweExample(final Activity activity,
                                       final String urlString,
                                       final String title) {
        GamebaseWebViewConfiguration configuration
                = new GamebaseWebViewConfiguration.Builder()
                .setStyle(GamebaseWebViewStyle.BROWSER)             // BROWSER, POPUP
                .setTitleText(title)                                // Set Title
                .setScreenOrientation(ScreenOrientation.PORTRAIT)   // PORTRAIT, LANDSCAPE, LANDSCAPE_SENSOR
                .setNavigationBarColor(Color.RED)                   // Set Navigation Bar Color
                .setNavigationBarHeight(40)                         // Set Navigation Bar Height
                .setBackButtonVisible(true)                         // Set Back Button Visible
//                .setBackButtonImageResource(R.id.back_button)       // Set Back Button Image ID of resource
//                .setCloseButtonImageResource(R.id.close_button)     // Set Close Button Image ID of resource
                .build();
        showWebView(activity, urlString, configuration, null, null, null);
    }

    private static void webviewExampleWithSchemeEvent(final Activity activity,
                                                      final String urlString,
                                                      final String title) {
        GamebaseWebViewConfiguration configuration
                = new GamebaseWebViewConfiguration.Builder()
                .setStyle(GamebaseWebViewStyle.BROWSER)             // BROWSER, POPUP
                .setTitleText(title)                                // Set Title
                .build();
        List<String> schemeList = new ArrayList<>();
        schemeList.add("mygame://");
        schemeList.add("closemywebview://");
        showWebView(activity, urlString, configuration,
                new GamebaseCallback() {
                    @Override
                    public void onCallback(GamebaseException exception) {
                        // When closed WebView, this callback will be called.
                    }
                },
                schemeList,
                new GamebaseDataCallback<String>() {
                    @Override
                    public void onCallback(String fullUrl, GamebaseException exception) {
                        if (Gamebase.isSuccess(exception)) {
                            if (fullUrl.contains("mygame://")) {
                                // do something
                            } else if (fullUrl.contains("closemywebview://")) {
                                closeWebView(activity);
                            }
                        } else {
                            AppFlowHelper.showError(activity, "Exception from WebView Callback", exception.toString());
                        }
                    }
                });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // GamebaseEventHandler
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void addGamebaseEventHandler(final Activity activity) {
        if (GAMEBASE_EVENT_HANDLER != null) {
            Gamebase.removeEventHandler(GAMEBASE_EVENT_HANDLER);
        }
        GAMEBASE_EVENT_HANDLER = message -> {
            switch (message.category) {
                case GamebaseEventCategory.LOGGED_OUT:
                    GamebaseEventLoggedOutData loggedOutData = GamebaseEventLoggedOutData.from(message.data);
                    if (loggedOutData != null) {
                        Log.i(TAG, "[GamebaseEventHandler] category : " + message.category);
                        Log.i(TAG, "[GamebaseEventHandler] loggedOutData : " + loggedOutData.toString());

                        Log.i(TAG, "[GamebaseEventHandler] Process Login again.");
                        Log.d(TAG, "--------------------------------------");
                        // There was a problem with the access token.
                        // Call login again.
                        Gamebase.login(activity, Gamebase.getLastLoggedInProvider(), (authToken, exception) -> {});
                    }
                    break;
                case GamebaseEventCategory.SERVER_PUSH_APP_KICKOUT_MESSAGE_RECEIVED:
                case GamebaseEventCategory.SERVER_PUSH_APP_KICKOUT:
                case GamebaseEventCategory.SERVER_PUSH_TRANSFER_KICKOUT:
                    GamebaseEventServerPushData serverPushData = GamebaseEventServerPushData.from(message.data);
                    if (serverPushData != null) {
                        processServerPush(activity, message.category, serverPushData);
                    }
                    break;
                case GamebaseEventCategory.OBSERVER_LAUNCHING:
                case GamebaseEventCategory.OBSERVER_HEARTBEAT:
                case GamebaseEventCategory.OBSERVER_NETWORK:
                    GamebaseEventObserverData observerData = GamebaseEventObserverData.from(message.data);
                    if (observerData != null) {
                        processObserver(activity, message.category, observerData);
                    }
                    break;
                case GamebaseEventCategory.PURCHASE_UPDATED:
                    PurchasableReceipt receipt = PurchasableReceipt.from(message.data);
                    if (receipt != null) {
                        // If the user got item by 'Promotion Code',
                        // this event will be occurred.
                        Log.i(TAG, "[GamebaseEventHandler] category : " + message.category);
                        Log.i(TAG, "[GamebaseEventHandler] PurchasableReceipt : " + receipt.toString());
                        Log.d(TAG, "--------------------------------------");
                    }
                    break;
                case GamebaseEventCategory.PUSH_RECEIVED_MESSAGE:
                    PushMessage pushMessage = PushMessage.from(message.data);
                    if (pushMessage != null) {
                        // When you received push message.
                        Log.i(TAG, "[GamebaseEventHandler] category : " + message.category);
                        Log.i(TAG, "[GamebaseEventHandler] PushMessage : " + pushMessage.toString());
                        Log.d(TAG, "--------------------------------------");
                        try {
                            JSONObject json = new JSONObject(pushMessage.extras);
                            // There is 'isForeground' information.
                            boolean isForeground = json.getBoolean("isForeground");
                            // You can get your custom key.
                            if (json.has("YourCustomKey")) {
                                Object customValue = json.get("YourCustomKey");
                            }
                        } catch (Exception ignored) {
                        }
                    }
                    break;
                case GamebaseEventCategory.PUSH_CLICK_MESSAGE:
                    PushMessage clickedMessage = PushMessage.from(message.data);
                    if (clickedMessage != null) {
                        // When you clicked push message.
                        Log.i(TAG, "[GamebaseEventHandler] category : " + message.category);
                        Log.i(TAG, "[GamebaseEventHandler] PushMessage(clickedMessage) : " + clickedMessage.toString());
                        Log.d(TAG, "--------------------------------------");
                    }
                    break;
                case GamebaseEventCategory.PUSH_CLICK_ACTION:
                    PushAction pushAction = PushAction.from(message.data);
                    if (pushAction != null) {
                        // When you clicked action button by 'Rich Message'.
                        Log.i(TAG, "[GamebaseEventHandler] category : " + message.category);
                        Log.i(TAG, "[GamebaseEventHandler] PushAction : " + pushAction.toString());
                        Log.d(TAG, "--------------------------------------");
                    }
                    break;
                default:
                    break;
            }
        };
        Gamebase.addEventHandler(GAMEBASE_EVENT_HANDLER);
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // ServerPush
    //
    ////////////////////////////////////////////////////////////////////////////////
    private static void processServerPush(final Activity activity,
                                          @NonNull final String category,
                                          @NonNull final GamebaseEventServerPushData data) {
        Log.d(TAG, "[GamebaseEventHandler] processServerPush");
        Log.i(TAG, "[GamebaseEventHandler] category : " + category);
        Log.i(TAG, "[GamebaseEventHandler] serverPushData : " + data.toString());
        Log.d(TAG, "--------------------------------------");

        if (category.equals(GamebaseEventCategory.SERVER_PUSH_APP_KICKOUT_MESSAGE_RECEIVED)) {
            // Currently, the kickout pop-up is displayed.
            // If your game is running, stop it.
            AppFlowHelper.pauseGame();
        } else if (category.equals(GamebaseEventCategory.SERVER_PUSH_APP_KICKOUT)) {
            // Kicked out from Gamebase server.(Maintenance, banned or etc..)
            // And the game user closes the kickout pop-up.
            // Return to title and initialize Gamebase again.
            AppFlowHelper.returnToTitle(activity);
        } else if (category.equals(GamebaseEventCategory.SERVER_PUSH_TRANSFER_KICKOUT)) {
            // If the user wants to move the guest account to another device,
            // if the account transfer is successful,
            // the login of the previous device is released,
            // so go back to the title and try to log in again.
            AppFlowHelper.returnToTitle(activity);
        }
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Observer
    //
    ////////////////////////////////////////////////////////////////////////////////
    private static void processObserver(final Activity activity,
                                        @NonNull final String category,
                                        @NonNull final GamebaseEventObserverData data) {
        Log.d(TAG, "[GamebaseEventHandler] processObserver");
        Log.i(TAG, "[GamebaseEventHandler] category : " + category);
        Log.i(TAG, "[GamebaseEventHandler] observerData : " + data.toString());
        Log.d(TAG, "--------------------------------------");

        if (category.equals(GamebaseEventCategory.OBSERVER_LAUNCHING)) {
            int launchingStatusCode = data.code;
            String messageString = data.message;
            Log.d(TAG, "Update launching status to " + launchingStatusCode + ", " + messageString);

            boolean canPlay = true;
            String errorLog = "";
            switch (launchingStatusCode) {
                case LaunchingStatus.IN_SERVICE:
                    break;
                case LaunchingStatus.RECOMMEND_UPDATE:
                    Log.d(TAG, "There is a new version of this application.");
                    break;
                case LaunchingStatus.IN_SERVICE_BY_QA_WHITE_LIST:
                case LaunchingStatus.IN_TEST:
                case LaunchingStatus.IN_REVIEW:
                case LaunchingStatus.IN_BETA:
                    Log.d(TAG, "You logged in because you are developer.");
                    break;
                case LaunchingStatus.REQUIRE_UPDATE:
                    canPlay = false;
                    errorLog = "You have to update this application.";
                    break;
                case LaunchingStatus.BLOCKED_USER:
                    canPlay = false;
                    errorLog = "You are blocked user!";
                    break;
                case LaunchingStatus.TERMINATED_SERVICE:
                    canPlay = false;
                    errorLog = "Game is closed!";
                    break;
                case LaunchingStatus.INSPECTING_SERVICE:
                case LaunchingStatus.INSPECTING_ALL_SERVICES:
                    canPlay = false;
                    errorLog = "Under maintenance.";
                    break;
                case LaunchingStatus.INTERNAL_SERVER_ERROR:
                default:
                    canPlay = false;
                    errorLog = "Unknown internal error.";
                    break;
            }
            if (canPlay) {
                // Now, user can play the game.
                Log.d(TAG, "Changed Status : You can play game, now!");
            } else {
                // User cannot play the game.
                Log.e(TAG, "Changed Status : " + errorLog);
                AppFlowHelper.returnToTitle(activity);
            }
        } else if (category.equals(GamebaseEventCategory.OBSERVER_HEARTBEAT)) {
            int errorCode = data.code;
            Log.d(TAG, "Heartbeat changing : " + data.toString());

            switch (errorCode) {
                case GamebaseError.INVALID_MEMBER:
                    // You can check the invalid user session in here.
                    // ex) After transferred account to another device.
                    break;
                case GamebaseError.BANNED_MEMBER:
                    // You can check the banned user session in here.
                    break;
            }
        } else if (category.equals(GamebaseEventCategory.OBSERVER_NETWORK)) {
            int networkTypeCode = data.code;
            Log.d(TAG, "Network changing : " + data.toString());

            // You can check the changed network status in here.
            if (networkTypeCode == NetworkManager.TYPE_NOT) {
                // Network disconnected.
            } else {
                // Network connected.
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Gamebase Analytics
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void initializeAnalytics(int userLevel, String channelId, String characterId) {
        GameUserData gameUserData = new GameUserData(userLevel);

        // Optional data
        gameUserData.channelId = channelId;     // Optional
        gameUserData.characterId = characterId; // Optional

        Gamebase.Analytics.setGameUserData(gameUserData);
    }

    public static void onGameUserLevelUp(int userLevel, long levelUpTime) {
        LevelUpData levelUpData = new LevelUpData(userLevel, levelUpTime);
        Gamebase.Analytics.traceLevelUp(levelUpData);
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // TOAST Logger (Logger&Crash)
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void initializeToastLogger(final Context context) {
        Log.d(TAG, "Initialize TOAST Logger");
        final LoggerConfiguration.Builder configBuilder = LoggerConfiguration.newBuilder(
                LOG_AND_CRASH_APPKEY, true);
        Gamebase.Logger.initialize(context, configBuilder.build());
    }

    public static void setToastLoggerListener() {
        Gamebase.Logger.setLoggerListener(new LoggerListener() {
            @Override
            public void onSuccess(@NotNull LogEntry logEntry) {
            }

            @Override
            public void onFilter(@NotNull LogEntry logEntry, @NotNull LogFilter logFilter) {
                Log.i(TAG, "SendLog.onFilter(" + logEntry.toString() + ") : " + logFilter.toString());
            }

            @Override
            public void onSave(@NotNull LogEntry logEntry) {
            }

            @Override
            public void onError(@NotNull LogEntry logEntry, @NotNull GamebaseException exception) {
                Log.w(TAG, "SendLog.onError(" + logEntry.toString() + ") : " + exception.toString());
            }
        });
    }

    public static void sendLogDebug(final String message, final Map<String, String> userField) {
        Gamebase.Logger.debug(message, userField);
    }

    public static void sendLogInfo(final String message, final Map<String, String> userField) {
        Gamebase.Logger.info(message, userField);
    }

    public static void sendLogWarn(final String message, final Map<String, String> userField) {
        Gamebase.Logger.warn(message, userField);
    }

    public static void sendLogError(final String message, final Map<String, String> userField) {
        Gamebase.Logger.error(message, userField);
    }

    public static void sendLogFatal(final String message, final Map<String, String> userField) {
        Gamebase.Logger.fatal(message, userField);
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // UI Helper
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void showSimpleAlertDialog(final Activity activity, final String title, final String message) {
        Gamebase.Util.showAlert(activity, title, message);
    }

    public static void showAlertDialogWithListener(final Activity activity,
                                                   final String title,
                                                   final String message,
                                                   final DialogInterface.OnClickListener clickListener) {
        Gamebase.Util.showAlert(activity, title, message, clickListener);
    }

    public static void showToast(final Activity activity, final String message, final int duration) {
        Gamebase.Util.showToast(activity, message, duration);
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Temporary Withdrawal
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void requestWithdrawal(final Activity activity, final GamebaseDataCallback<TemporaryWithdrawalInfo> callback) {
        Gamebase.TemporaryWithdrawal.requestWithdrawal(activity, new GamebaseDataCallback<TemporaryWithdrawalInfo>() {
            @Override
            public void onCallback(TemporaryWithdrawalInfo temporaryWithdrawalInfo, GamebaseException exception) {
                if (callback != null) {
                    callback.onCallback(temporaryWithdrawalInfo, exception);
                }
            }
        });
    }

    public static void cancelWithdrawal(final Activity activity, final GamebaseCallback callback) {
        Gamebase.TemporaryWithdrawal.cancelWithdrawal(activity, new GamebaseCallback() {
            @Override
            public void onCallback(GamebaseException exception) {
                if (callback != null) {
                    callback.onCallback(exception);
                }
            }
        });
    }

    public static void withdrawImmediately(final Activity activity, final GamebaseCallback callback) {
        Gamebase.TemporaryWithdrawal.withdrawImmediately(activity, new GamebaseCallback() {
            @Override
            public void onCallback(GamebaseException exception) {
                if (callback != null) {
                    callback.onCallback(exception);
                }
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Contact
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void openContact(final Activity activity, final String userName, final Callback callback) {
        ContactConfiguration.Builder builder = ContactConfiguration.newBuilder();
        if (userName != null) {
            builder.setUserName(userName);
        }

        Gamebase.Contact.openContact(activity, builder.build(), new GamebaseCallback() {
            @Override
            public void onCallback(GamebaseException exception) {
                if (callback != null) callback.onCallback();
            }
        });
    }

    public static void requestContactURL(final String userName, final GamebaseDataCallback<String> callback) {
        ContactConfiguration.Builder builder = ContactConfiguration.newBuilder();
        if (userName != null) {
            builder.setUserName(userName);
        }

        Gamebase.Contact.requestContactURL(builder.build(), new GamebaseDataCallback<String>() {
            @Override
            public void onCallback(String contactUrl, GamebaseException exception) {
                if (callback != null) callback.onCallback(contactUrl, exception);
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Terms
    //
    ////////////////////////////////////////////////////////////////////////////////
    public static void showTermsView(final Activity activity,
                                     final GamebaseDataCallback<GamebaseDataContainer> callback) {
        Gamebase.Terms.showTermsView(activity, new GamebaseDataCallback<GamebaseDataContainer>() {
            @Override
            public void onCallback(GamebaseDataContainer container, GamebaseException exception) {
                if (callback != null) callback.onCallback(container, exception);
            }
        });
    }

    public static void queryTerms(final Activity activity,
                                  final GamebaseDataCallback<GamebaseQueryTermsResult> callback) {
        Gamebase.Terms.queryTerms(activity, new GamebaseDataCallback<GamebaseQueryTermsResult>() {
            @Override
            public void onCallback(GamebaseQueryTermsResult gamebaseQueryTermsResult, GamebaseException exception) {
                if (Gamebase.isSuccess(exception)) {
                    // Succeeded.
                    final int termsSeq = gamebaseQueryTermsResult.getTermsSeq();
                    final String termsVersion = gamebaseQueryTermsResult.getTermsVersion();
                    final String termsCountryType = gamebaseQueryTermsResult.getTermsCountryType();
                    final List<GamebaseTermsContentDetail> contents = gamebaseQueryTermsResult.getContents();
                } else if (exception.getCode() == GamebaseError.UI_TERMS_NOT_EXIST_FOR_DEVICE_COUNTRY) {
                    // Another country device.
                    // Pass the 'terms and conditions' step.
                } else {
                    // Failed.
                }

                if (callback != null) callback.onCallback(gamebaseQueryTermsResult, exception);
            }
        });
    }

    public static void updateTerms(final Activity activity,
                                   final GamebaseUpdateTermsConfiguration configuration,
                                   final GamebaseCallback callback) {
        Gamebase.Terms.updateTerms(activity, configuration, new GamebaseCallback() {
            @Override
            public void onCallback(GamebaseException exception) {
                if (callback != null) callback.onCallback(exception);
            }
        });
    }

    public static void updateTermsExample(final Activity activity,
                                          final GamebaseCallback callback) {
        queryTerms(activity, new GamebaseDataCallback<GamebaseQueryTermsResult>() {
            @Override
            public void onCallback(final GamebaseQueryTermsResult queryTermsResult,
                                   final GamebaseException queryTermsException) {
                final int termsSeq = queryTermsResult.getTermsSeq();
                final String termsVersion = queryTermsResult.getTermsVersion();
                final List<GamebaseTermsContent> contents = new ArrayList<>();
                for (GamebaseTermsContentDetail detail : queryTermsResult.getContents()) {
                    GamebaseTermsContent content = GamebaseTermsContent.from(detail);
                    if (content != null) {
                        // Change 'agreed' value for test.
                        content.setAgreed(!content.getAgreed());
                        contents.add(content);
                    }
                }
                final GamebaseUpdateTermsConfiguration configuration =
                        GamebaseUpdateTermsConfiguration.newBuilder(termsSeq, termsVersion, contents)
                                .build();
                updateTerms(activity, configuration, new GamebaseCallback() {
                    @Override
                    public void onCallback(GamebaseException updateTermsException) {
                        if (callback != null) {
                            callback.onCallback(updateTermsException);
                        }
                    }
                });
            }
        });
    }

    ////////////////////////////////////////////////////////////////////////////////
    //
    // Internal Callback
    //
    ////////////////////////////////////////////////////////////////////////////////

    public interface Callback {
        void onCallback();
    }
}
